﻿-- Implemented by MrCool (EU, FR, Hyjal, main character: Eilylune)

local data = PVPTooltips_Data;
local secondaryData = PVPTooltips_SecondaryData;
local itemData = PVPTooltips_ItemData;
local redirection = PVPTooltips_Redirection;
local patch = PVPTooltips_Patch[GetLocale():sub(1, 2)] or {};
local history = PVPTooltips_History;
local version = GetAddOnMetadata("PVPTooltips", "Version");

-- PART 1 - Tooltips

-- Create array of specialization IDs

local SPECIALIZATIONS = { 250, 251, 252, 577, 581, 102, 103, 104, 105, 253, 254, 255, 62, 63, 64, 268, 270, 269, 65, 66, 70, 256, 257, 258, 259, 260, 261, 262, 263, 264, 265, 266, 267, 71, 72, 73 };
local MATCHES = { };

local function EvaluateSpecFilter(spec)
	local isMatch, filterText = true, "";
	if spec then
		-- Compute filter text
		wipe(MATCHES);
		local id, name, _, _, _, class = GetSpecializationInfoByID(abs(spec));
		if spec > 0 then
			tinsert(MATCHES, name);
		else
			for _, otherId in ipairs(SPECIALIZATIONS) do
				local _, otherName, _, _, _, otherClass = GetSpecializationInfoByID(otherId);
				if otherId ~= id and class == otherClass then
					tinsert(MATCHES, otherName);
				end
			end
		end
		filterText = " (" .. table.concat(MATCHES, ", ") .. ")";
		
		-- Evaluate if current spec is matching
		local currentIndex = GetSpecialization();
		local currentId = GetSpecializationInfo(currentIndex or 1);
		isMatch = (spec > 0 and currentId == spec) or (spec < 0 and currentId ~= -spec);
	end
	return isMatch, filterText;
end

-- Spell modifiers access 

local MODIFIERS = { };

local function GetModifier(id, dataset)
	id = id or 0
	id = redirection[id] or id
	
	if dataset == 1 then
		return data[id];
	elseif dataset == 2 then
		return secondaryData[id];
	else
		return nil;
	end
end

local function GetModifiers(id)
	wipe(MODIFIERS);
	local info = GetModifier(id, 1);
	if info then
		tinsert(MODIFIERS, info);
	end
	info = GetModifier(id, 2);
	if info then
		tinsert(MODIFIERS, info);
	end
	return MODIFIERS;
end

-- We need to do some annoying things to properly handle tooltip modification:
--	  Remove color text from the tooltip (occurs for artifacts) then restore it
--    Remove % and thousand separator
--    Preserve decimal separator and replace it by a dot "." so we can parse decimal numbers
--    Calculate the modified value and replace it in the original text by the localized number
--    We also need to handle small values with a decimal part and rounding errors

local TRANSFORM = { };
local KEYS = {"__a__", "__b__", "__c__", "__d__", "__e__"}; -- Should be more than enough
local PATCH = { }; -- For efficiency

local function Escape(text)
	wipe(TRANSFORM);
	text = text:gsub("|c%x%x%x%x%x%x%x%x.+|r", function(x)
		tinsert(TRANSFORM, x);
		return KEYS[#TRANSFORM];
	end);
	return text;
end

local function Unescape(text)
	for i=1, #TRANSFORM do
		text = text:gsub(KEYS[i], TRANSFORM[i]);
	end
	wipe(TRANSFORM);
	return text;
end

local function GetDecimalSeparator()
	 -- Contains no-break space (right blank)
	if IsEuropeanNumbers() then
		return "%,", "[  %%]"; -- NOTE: dot is sometimes used as decimal on some tooltips
	else
		return "%.", "[  ,%%]";
	end
end

local function Round(value, decimal)
	local d = 10 ^ decimal;
	return floor((value * d) + 0.5) / d;
end

local function WrapColor(text, multiplier)
	if multiplier >= 0 then
		if multiplier == 1 then
			return text;
		elseif multiplier < 1 then
			return "|cffff0000" .. text .. "|r";
		else
			return "|cff00ff00" .. text .. "|r";
		end
	else
		if multiplier == -1 then
			return text;
		elseif multiplier < -1 then
			return "|cffff0000" .. text .. "|r";
		else
			return "|cff00ff00" .. text .. "|r";
		end
	end
end

local function ReplaceNumber(x, value)
	local s, e;
	for i=1, #x do
		local code = x:byte(i);
		if code >= 48 and code <= 57 then
			s = s or i;
			e = i;
		end
	end
	local decimal = 0;
	if value < 1000 then decimal = 2; end
	return x:sub(1, s-1) .. BreakUpLargeNumbers(Round(value, decimal)) .. x:sub(e+1);
end

local function TransformValueText(x, multiplier)
	local decimal, pattern = GetDecimalSeparator();
	local value = x:gsub(pattern, ""):gsub(decimal, ".");
	local replacement = ReplaceNumber(x, tonumber(value) * abs(multiplier));
	--print(x, value, replacement);
	return WrapColor(replacement, multiplier);
end

local function ApplyMultiplier(patch)
	local current = 0
	return function(x)
		current = current + 1
		--print(x, current)
		for i=1, #patch, 2 do
			local position = patch[i];
			local multiplier = patch[i+1];
			
			if type(position) == "table" then
				if tContains(position, current) then
					return TransformValueText(x, multiplier);
				end
			else
				if current == position then
					return TransformValueText(x, multiplier);
				end
			end
		end
		return x
	end
end

local function PatchTooltip(patch)
	for i=1, select("#", GameTooltip:GetRegions()) do
		local region = select(i, GameTooltip:GetRegions());
		if region and region:GetObjectType() == "FontString" then
			local text = region:GetText();
			local r, g, b = region:GetTextColor();
			
			-- Identify spell description text
			if text and abs(r - 1.00) < 0.01 and abs(g - 0.82) < 0.01 and abs(b - 0.00) < 0.01 then
				--print(text);
				text = Escape(text);
				text = text:gsub("(%d[%d  ,.]*%%?)", ApplyMultiplier(patch)); -- Contains no-break space (right blank)
				text = Unescape(text);
				region:SetText(text);
			end
		end
	end
end

local function GetSuffix(filterText, details)
	local suffix = "";
	if details then suffix = " |cffffff00[" .. GetSpellInfo(details) .. "]|r"; end
	if filterText then suffix = suffix .. filterText; end
	return suffix;
end

local function UpdateTooltip(id)
	wipe(PATCH);
	
	-- If spell is unspecified, use current spell on tooltip
	if not id then
		local _, spellId = GameTooltip:GetSpell();
		id = spellId;
	end

	-- Append PVP multipliers
	for _, info in ipairs(GetModifiers(id)) do
		local position = patch[id] or info[1];
		local multiplier = info[2];
		local isMatch, filterText = EvaluateSpecFilter(info[3]);

		-- Verify we are in the correct specialization for a PVP modifier to take effect
		if isMatch and multiplier ~= 1.00 then
			if IsAltKeyDown() and not UnitIsPVP("player") and not UnitIsPVPFreeForAll("player") then
				tinsert(PATCH, position);
				tinsert(PATCH, multiplier);
			end
			GameTooltip:AddLine(PVP .. HEADER_COLON .. WrapColor((" %d%%"):format((abs(multiplier) * 100) + 0.5), multiplier) .. GetSuffix(filterText, info[4]), 1, 1, 1);
			GameTooltip:AddTexture("Interface\\ICONS\\Ability_DualWield");
		end
	end
	
	if #PATCH > 0 then
		-- Patch the tooltip values
		PatchTooltip(PATCH);
	end
	
	-- Refresh tooltip
	GameTooltip:Show();
end

local function UpdateItemTooltip(id)
	-- If item is unspecified, use current item on tooltip
	if not id then
		local _, itemLink = GameTooltip:GetItem();
		if itemLink then
			id = GetItemInfoInstant(itemLink);
		end
	end

	-- Append PVP multiplier
	local multiplier = itemData[id or 0] or 1;
	if multiplier ~= 1.00 then
		GameTooltip:AddLine(PVP .. HEADER_COLON .. WrapColor((" %d%%"):format((abs(multiplier) * 100) + 0.5), multiplier), 1, 1, 1);
		GameTooltip:AddTexture("Interface\\ICONS\\Ability_DualWield");
		GameTooltip:Show();
	end
end

do
	-- Setup required hooks for spells
	hooksecurefunc(GameTooltip, "SetAction", function() UpdateTooltip(); end);
	hooksecurefunc(GameTooltip, "SetSpellBookItem", function() UpdateTooltip(); end);
	hooksecurefunc(GameTooltip, "SetSpellByID", function() UpdateTooltip(); end);
	hooksecurefunc(GameTooltip, "SetTalent", function() UpdateTooltip(); end);
	hooksecurefunc(GameTooltip, "SetShapeshift", function() UpdateTooltip(); end);
	hooksecurefunc(GameTooltip, "SetAzeritePower", function() UpdateTooltip(); end);

	-- Setup required hooks for items
	hooksecurefunc(GameTooltip, "SetBagItem", function() UpdateItemTooltip(); end);
	hooksecurefunc(GameTooltip, "SetInventoryItem", function() UpdateItemTooltip(); end);
	--hooksecurefunc(GameTooltip, "SetInventoryItemByID", function() UpdateItemTooltip(); end);
	--hooksecurefunc(GameTooltip, "SetItemByID", function() UpdateItemTooltip(); end);
end

-- PART 2 - History

local DEFAULT_COLOR = {r = 1, g = 1, b = 1};

local function Write(msg)
	if DEFAULT_CHAT_FRAME then
		DEFAULT_CHAT_FRAME:AddMessage(msg);
	else
		print(msg);
	end
end

local function GetDeltaString(old, new)
	local newStr = ("%d%%"):format(abs(new * 100) + 0.5);
	if new == old then
		return newStr;
	else
		local oldStr = ("%d%%"):format(abs(old * 100) + 0.5);
		return WrapColor(oldStr .. " → " .. newStr, abs(new) / abs(old) * (new < 0 and -1 or 1));
	end
end

local function WriteDiff(version)
	local spellChanges = history.Spells[version];

	if spellChanges then
		for _, change in ipairs(spellChanges) do
			local class = change[1];
			local spell = change[2];
			local old = change[3];
			local new = change[4];
			local dataset = change[5] or 1;
			local name, _, icon = GetSpellInfo(spell);
			
			if name then
				local info = GetModifier(spell, dataset);
				local _, filterText = EvaluateSpecFilter(info[3]);
				local color = RAID_CLASS_COLORS[class] or DEFAULT_COLOR;
				local spellIcon = ("|T%s.blp:0|t"):format(icon);
				local spellLink = (" |Hspell:%d|h|cff%02x%02x%02x[%s]|r|h"):format(spell, floor(color.r * 255), floor(color.g * 255), floor(color.b * 255), name);
					
				Write(spellIcon .. spellLink .. HEADER_COLON .. " " .. GetDeltaString(old, new) .. GetSuffix(filterText, info[4]));
			end
		end
	end
end

function PVPTooltips_ShowChanges(v)
	WriteDiff(v or version);
end

do
	Write("|cffffff20PVP Tooltips (" .. version .. ")|r");
	WriteDiff(version);
end